/*
 * Proxy/Stub implementation
 *
 * Copyright 2001 Ove Kven, TransGaming Technologies
 *
 * (Consider this file a large FIXME until stuff is done more thoroughly,
 *  many many things need to move from here to dlls/rpcrt4)
 *
 */

#include <string.h>

#include "windef.h"
#include "winbase.h"
#include "winerror.h"
#include "ole2.h"

#include "wine/obj_base.h"
#include "wine/obj_storage.h"
#include "wine/obj_marshal.h"
#include "wine/obj_channel.h"

#include "debugtools.h"

DEFAULT_DEBUG_CHANNEL(ole);

/* I think that the standard marshaller returns a special GUID to
 * signal standard proxy/stub-based marshaling; I don't know exactly
 * which one, but I picked up this from Win2000's registry, which
 * I assume means "Default Marshaler", so I just use that for now */
static GUID CLSID_DfMarshal = {
  0x30B, 0, 0, {0xC0, 0, 0, 0, 0, 0, 0, 0x46}
};

/* COM_GetPSFactory [internal]:
 * Given an interface ID riid, this function returns an IPSFactoryBuffer
 * interface which can be used to create marshalling proxies or stubs for
 * the given interface. */
static HRESULT COM_GetPSFactory(REFIID riid, LPPSFACTORYBUFFER *pPS)
{
  HRESULT hr;
  CLSID clsid;
  hr = CoGetPSClsid(riid, &clsid);
  if (FAILED(hr)) return hr;
  hr = CoGetClassObject(&clsid, CLSCTX_INPROC_SERVER, NULL,
			&IID_IPSFactoryBuffer, (LPVOID *)pPS);
  return hr;
}

/* COM tracks interfaces with three kinds of IDs:
 *  OXID = Object Exporter ID, one for each apartment (unique within network)
 *  OID = Object ID, one for each object (unique within apartment)
 *  IPID = Interface Pointer ID, one for each interface (unique within apartment)
 *
 * Hence, each "stub manager" would correspond to an OID, and each
 * of the interface stubs would correspond to an IPID.
 */

/******************** STUB MANAGER ********************/

/* FIXME: this should be tracked per apartment, not per process,
 * and also certainly not in simple-minded static tables like this... */

static struct OID_entry {
  LPUNKNOWN pUnk;        /* object identity */
  /* FIXME: pointer to stub manager here */
} OID_table[256];

static struct IPID_entry {
  DWORD OID;             /* OID of stub manager */
  IID iid;               /* IID of interface */
  LPVOID pv;             /* interface pointer */
  LPRPCSTUBBUFFER pStub; /* interface stub */
  DWORD ref;             /* external reference count */
  HRESULT hres;          /* result of stub creation */
} IPID_table[256];

static DWORD COM_GetOXID(void)
{
  /* FIXME: this is rotten... OXIDs should be per-apartment,
   * not per-process, and each OXID should also be unique within
   * the network (for DCOM) */
  return GetCurrentProcessId();
}

/* FIXME: thread safety, really lousy design, etc... */
static HRESULT COM_GetID(REFIID riid, LPVOID pv, DWORD *OID, DWORD *IPID)
{
  HRESULT hr;
  LPUNKNOWN pUnk;
  LPPSFACTORYBUFFER pPSFactory;
  int i, fn;
  /* see if the interface is already there */
  for (fn=-1,i=0; i<sizeof(IPID_table)/sizeof(struct IPID_entry); i++) {
    if (IPID_table[i].pv == pv &&
	IsEqualGUID(riid,&IPID_table[i].iid)) {
      *OID = IPID_table[i].OID;
      *IPID = i;
      return IPID_table[i].hres;
    }
    /* find free spot */
    if (fn==-1 && !IPID_table[i].pv) fn = i;
  }
  /* need to add interface */
  if (fn == -1) {
    FIXME("IPID table is out of space!\n");
    return E_OUTOFMEMORY;
  }
  /* ensure that we have the object identity */
  hr = IUnknown_QueryInterface((LPUNKNOWN)pv, &IID_IUnknown, (LPVOID*)&pUnk);
  if (FAILED(hr)) return hr;
  IUnknown_Release(pUnk);
  /* add interface */
  *IPID = fn;
  IPID_table[fn].pv = pv;
  memcpy(&IPID_table[fn].iid,riid,sizeof(IPID_table[fn].iid));
  /* create interface stub */
  IPID_table[fn].hres = COM_GetPSFactory(riid, &pPSFactory);
  if (SUCCEEDED(IPID_table[fn].hres)) {
    IPID_table[fn].hres = IPSFactoryBuffer_CreateStub(pPSFactory, riid, pv,
						      &IPID_table[*IPID].pStub);
    IPSFactoryBuffer_Release(pPSFactory);
  }
  /* check for object identity */
  for (fn=-1,i=0; i<sizeof(OID_table)/sizeof(struct OID_entry); i++) {
    if (OID_table[i].pUnk == pUnk) {
      *OID = i;
      IPID_table[*IPID].OID = i;
      return S_OK;
    }
    /* find free spot */
    if (fn==-1 && !OID_table[i].pUnk) fn = i;
  }
  /* need to add object */
  if (fn == -1) {
    FIXME("OID table is out of space!\n");
    return E_OUTOFMEMORY;
  }
  OID_table[fn].pUnk = pUnk;
  *OID = fn;
  IPID_table[*IPID].OID = fn;
  /* success */
  return IPID_table[*IPID].hres;
}

HRESULT COM_GetStub(DWORD OXID,DWORD OID,DWORD IPID,LPRPCSTUBBUFFER *ppStub)
{
  *ppStub = IPID_table[IPID].pStub;
  if (!*ppStub) return E_FAIL;
  return S_OK;
}

void COM_KillStubs(void)
{
  int i;
  for (i=sizeof(IPID_table)/sizeof(struct IPID_entry); i>0; ) {
    i--;
    if (IPID_table[i].pv && IPID_table[i].pStub) {
      IRpcStubBuffer_Release(IPID_table[i].pStub);
      IPID_table[i].pv = NULL;
    }
  }
  for (i=sizeof(OID_table)/sizeof(struct OID_entry); i>0; ) {
    i--;
    if (OID_table[i].pUnk)
      OID_table[i].pUnk = NULL;
  }
}

/******************** PROXY MANAGER ********************/

HRESULT COM_ConnectProxy(LPVOID pv, LPRPCCHANNELBUFFER pChannel);

typedef struct ProxyInfo {
  struct ProxyInfo *next;
  DWORD IPID;              /* IPID of interface */
  IID iid;                 /* IID of interface */
  LPRPCPROXYBUFFER pProxy; /* interface proxy */
  LPVOID pv;               /* interface pointer */
  DWORD owned;             /* owned external references */
  HRESULT hres;            /* result of proxy creation */
} ProxyInfo;

typedef struct ProxyManImpl {
  ICOM_VTABLE(IUnknown) *lpVtbl;
  DWORD ref;               /* local reference count */
  DWORD OXID;              /* OXID of object's apartment */
  DWORD OID;               /* OID of object's stub manager */
  ProxyInfo *info;         /* interfaces */
  LPRPCCHANNELBUFFER pChannel;
} ProxyManImpl;

static HRESULT WINAPI ProxyMan_QueryInterface(LPUNKNOWN iface,
					      REFIID riid,
					      LPVOID *obj)
{
  ICOM_THIS(ProxyManImpl,iface);
  TRACE("(%p)->QueryInterface(%s,%p)\n",This,debugstr_guid(riid),obj);
  if (IsEqualGUID(&IID_IUnknown,riid)) {
    *obj = This;
    This->ref++;
    return S_OK;
  }
  else {
    ProxyInfo *info = This->info;
    while (info) {
      if (IsEqualGUID(&info->iid,riid)) {
        *obj = info->pv;
        IUnknown_AddRef((LPUNKNOWN)*obj);
        return S_OK;
      }
      info = info->next;
    }
    /* if not found locally, ask server for interface
     * (should use IRemUnknown, but we don't yet) */
{
  RPCOLEMESSAGE msg;
  HRESULT hr;
  ULONG status = E_NOINTERFACE;
  TRACE("(%p)->RemoteQueryInterface(%s,%p)\n",This,debugstr_guid(riid),obj);
  msg.reserved1 = This; /* BAD hack */
  msg.iMethod = (ULONG)-1;
  msg.cbBuffer = sizeof(*riid);
  msg.Buffer = NULL;
  IRpcChannelBuffer_GetBuffer(This->pChannel, &msg, riid);
  memcpy(msg.Buffer, riid, sizeof(*riid));
  hr = IRpcChannelBuffer_SendReceive(This->pChannel, &msg, &status);
  if (FAILED(hr)) { TRACE("failed: %08lx\n", hr); *obj = NULL; return hr; }
  if (msg.cbBuffer) {
    LPSTREAM pStm = NULL;
    ULARGE_INTEGER cbSize;
    HGLOBAL hGlobal = 0;
    hr = CreateStreamOnHGlobal(0, TRUE, &pStm);
    cbSize.s.HighPart = 0;
    cbSize.s.LowPart = msg.cbBuffer;
    IStream_SetSize(pStm, cbSize);
    GetHGlobalFromStream(pStm, &hGlobal);
    memcpy(GlobalLock(hGlobal), msg.Buffer, msg.cbBuffer);
    GlobalUnlock(hGlobal);
    hr = CoUnmarshalInterface(pStm, riid, obj);
    IStream_Release(pStm);
    /* HACK: make sure the new proxy is connected */
    if (SUCCEEDED(hr)) {
      COM_ConnectProxy(*obj, This->pChannel);
      TRACE("unmarshalled => %p\n", *obj);
    } else {
      ERR("unmarshal failed: %lx\n", hr);
    }
    return hr;
  } else { TRACE("no reply\n"); *obj = NULL; return E_NOINTERFACE; }
}
  }
  return E_NOINTERFACE;
}

static ULONG WINAPI ProxyMan_AddRef(LPUNKNOWN iface)
{
  ICOM_THIS(ProxyManImpl,iface);
  TRACE("(%p)->AddRef()\n",This);
  return ++(This->ref);
}

static ULONG WINAPI ProxyMan_Release(LPUNKNOWN iface)
{
  ICOM_THIS(ProxyManImpl,iface);
  TRACE("(%p)->Release()\n",This);
  if (!--(This->ref)) {
    /* unneeded proxies should probably be cleaned up someday... */
    return 0;
  }
  return This->ref;
}

static ICOM_VTABLE(IUnknown) ProxyMan_VTable =
{
  ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
  ProxyMan_QueryInterface,
  ProxyMan_AddRef,
  ProxyMan_Release
};

/* FIXME: this should be tracked per apartment, not per process,
 * and also certainly not in simple-minded static tables like this... */
static struct proxy_entry {
  ProxyManImpl *proxy; /* local proxy */
} proxy_table[256];

/* FIXME: thread safety, really lousy design, etc... */
static HRESULT COM_GetProxy(DWORD OXID, DWORD OID, DWORD IPID, REFIID riid, LPVOID *ppv)
{
  ProxyManImpl *proxy = NULL;
  ProxyInfo *info = NULL;
  LPPSFACTORYBUFFER pPSFactory;
  int i, fn;
  for (fn=-1,i=0; i<sizeof(proxy_table)/sizeof(struct proxy_entry); i++) {
    if (proxy_table[i].proxy &&
	proxy_table[i].proxy->OXID == OXID &&
	proxy_table[i].proxy->OID == OID) {
      proxy = proxy_table[i].proxy;
      break;
    }
    /* find free spot */
    if (fn==-1 && !proxy_table[i].proxy) fn = i;
  }
  if (!proxy) {
    /* need to add proxy manager */
    if (fn == -1) {
      FIXME("proxy table is out of space!\n");
      return E_OUTOFMEMORY;
    }
    if (!riid) {
      FIXME("proxy not registered\n");
      return E_FAIL;
    }
    proxy = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(ProxyManImpl));
    proxy->lpVtbl = &ProxyMan_VTable;
    proxy->OXID = OXID;
    proxy->OID = OID;
    proxy_table[fn].proxy = proxy;
  }
  info = proxy->info;
  while (info && info->IPID != IPID) info = info->next;
  if (!info) {
    /* need to add interface */
    if (!riid) {
      FIXME("proxy not registered\n");
      return E_FAIL;
    }
    info = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(ProxyInfo));
    if (!info) return E_OUTOFMEMORY;
    info->next = proxy->info;
    info->IPID = IPID;
    memcpy(&info->iid, riid, sizeof(info->iid));
    proxy->info = info;
    /* create interface proxy (aggregated into proxy manager) */
    info->hres = COM_GetPSFactory(riid, &pPSFactory);
    if (FAILED(info->hres)) return info->hres;
    info->hres = IPSFactoryBuffer_CreateProxy(pPSFactory, (LPUNKNOWN)proxy, riid,
					      &info->pProxy, &info->pv);
    IPSFactoryBuffer_Release(pPSFactory);
    if (FAILED(info->hres)) return info->hres;
  }
  else {
    if (!info->pv) return info->hres;
  }
  /* success */
  *ppv = info->pv;
  return S_OK;
}

HRESULT COM_FindProxy(LPVOID pv, DWORD*OXID, DWORD*OID, DWORD*IPID, LPRPCPROXYBUFFER*pProxy)
{
  ProxyManImpl *proxy = NULL;
  ProxyInfo *info;
  /* get main object */
  IUnknown_QueryInterface((LPUNKNOWN)pv,&IID_IUnknown,(LPVOID*)&proxy);
  IUnknown_Release((LPUNKNOWN)proxy);
  /* is it a proxy? */
  if (proxy->lpVtbl == &ProxyMan_VTable) {
    /* it is, find the IDs */
    if (OXID) *OXID = proxy->OXID;
    if (OID) *OID = proxy->OID;
    /* find the requested interface */
    info = proxy->info;
    while (info) {
      if (info->pv == pv || proxy == pv) {
	if (IPID) *IPID = info->IPID;
	if (pProxy) *pProxy = info->pProxy;
	return S_OK;
      }
      info = info->next;
    }
    return E_FAIL;
  }
  /* it's not a proxy */
  if (pProxy) *pProxy = NULL;
  return S_FALSE;
}

HRESULT COM_ConnectProxy(LPVOID pv, LPRPCCHANNELBUFFER pChannel)
{
  LPRPCPROXYBUFFER pProxy = NULL;
  HRESULT hr = COM_FindProxy(pv, NULL, NULL, NULL, &pProxy);
  if (hr == S_OK) {
    ProxyManImpl *pman = NULL;
    /* get main object */
    IUnknown_QueryInterface((LPUNKNOWN)pv,&IID_IUnknown,(LPVOID*)&pman);
    IUnknown_Release((LPUNKNOWN)pman);
    TRACE("%p => %p, %p\n", pChannel, pProxy, pman);
    pman->pChannel = pChannel;
    return IRpcProxyBuffer_Connect(pProxy, pChannel);
  }
  return hr;
}

/******************** STANDARD MARSHALER ********************/

typedef struct StdMarshalImpl {
  ICOM_VTABLE(IMarshal) *lpVtbl;
  DWORD ref;
} StdMarshalImpl;

static HRESULT WINAPI StdMarshal_QueryInterface(LPMARSHAL iface,
						REFIID riid,
						LPVOID *obj)
{
  ICOM_THIS(StdMarshalImpl,iface);
  TRACE("(%p)->QueryInterface(%s,%p)\n",This,debugstr_guid(riid),obj);
  if (IsEqualGUID(&IID_IUnknown,riid) ||
      IsEqualGUID(&IID_IMarshal,riid)) {
    *obj = This;
    This->ref++;
    return S_OK;
  }
  return E_NOINTERFACE;
}

static ULONG WINAPI StdMarshal_AddRef(LPMARSHAL iface)
{
  ICOM_THIS(StdMarshalImpl,iface);
  TRACE("(%p)->AddRef()\n",This);
  return ++(This->ref);
}

static ULONG WINAPI StdMarshal_Release(LPMARSHAL iface)
{
  ICOM_THIS(StdMarshalImpl,iface);
  TRACE("(%p)->Release()\n",This);
  if (!--(This->ref)) {
    HeapFree(GetProcessHeap(),0,This);
    return 0;
  }
  return This->ref;
}

static HRESULT WINAPI StdMarshal_GetUnmarshalClass(LPMARSHAL iface,
						   REFIID riid,
						   LPVOID pv,
						   DWORD dwDestContext,
						   LPVOID pvDestContext,
						   DWORD mshlFlags,
						   CLSID *pCid)
{
  ICOM_THIS(StdMarshalImpl,iface);
  TRACE("(%p)->GetUnmarshalClass(%s,%p,%lx,%p,%lx,%p)\n",This,
	debugstr_guid(riid),pv,dwDestContext,pvDestContext,mshlFlags,pCid);
  /* return CLSID of the standard marshaller */
  memcpy(pCid, &CLSID_DfMarshal,sizeof(*pCid));
  return S_OK;
}

static HRESULT WINAPI StdMarshal_GetMarshalSizeMax(LPMARSHAL iface,
						   REFIID riid,
						   LPVOID pv,
						   DWORD dwDestContext,
						   LPVOID pvDestContext,
						   DWORD mshlFlags,
						   DWORD *pSize)
{
  ICOM_THIS(StdMarshalImpl,iface);
  TRACE("(%p)->GetMarshalSizeMax(%s,%p,%lx,%p,%lx,%p)\n",This,
	debugstr_guid(riid),pv,dwDestContext,pvDestContext,mshlFlags,pSize);
  /* marshal all kinds of IDs */
  *pSize = sizeof(DWORD)*3;
  return S_OK;
}

static HRESULT WINAPI StdMarshal_MarshalInterface(LPMARSHAL iface,
						  LPSTREAM pStm,
						  REFIID riid,
						  LPVOID pv,
						  DWORD dwDestContext,
						  LPVOID pvDestContext,
						  DWORD mshlFlags)
{
  ICOM_THIS(StdMarshalImpl,iface);
  DWORD OXID, OID, IPID;
  HRESULT hr;
  TRACE("(%p)->MarshalInterface(%p,%s,%p,%lx,%p,%lx)\n",This,pStm,
	debugstr_guid(riid),pv,dwDestContext,pvDestContext,mshlFlags);
  /* register interface (creating interface stubs as necessary),
   * and get lots of IDs in return */
  OXID = COM_GetOXID();
  hr = COM_GetID(riid, pv, &OID, &IPID);
  if (FAILED(hr)) return hr;
  TRACE("OXID=%ld, OID=%ld, IPID=%ld\n", OXID, OID, IPID);
  /* marshal those IDs (FIXME: use MEOW structure?)
   * FIXME: add error checking */
  IStream_Write(pStm, &OXID, sizeof(OXID), NULL);
  IStream_Write(pStm, &OID, sizeof(OID), NULL);
  IStream_Write(pStm, &IPID, sizeof(IPID), NULL);
  return S_OK;
}

static HRESULT WINAPI StdMarshal_UnmarshalInterface(LPMARSHAL iface,
						    LPSTREAM pStm,
						    REFIID riid,
						    LPVOID *ppv)
{
  ICOM_THIS(StdMarshalImpl,iface);
  DWORD OXID, OID, IPID;
  TRACE("(%p)->UnmarshalInterface(%p,%s,%p)\n",This,pStm,
	debugstr_guid(riid),ppv);
  /* unmarshal lots of IDs (FIXME: use MEOW structure?)
   * FIXME: add error checking */
  IStream_Read(pStm, &OXID, sizeof(OXID), NULL);
  IStream_Read(pStm, &OID, sizeof(OID), NULL);
  IStream_Read(pStm, &IPID, sizeof(IPID), NULL);
  TRACE("OXID=%ld, OID=%ld, IPID=%ld\n", OXID, OID, IPID);
  if (OXID == COM_GetOXID()) {
    /* dude, we're still in the same apartment...
     * we don't need no stinkin' marshaller */
    *ppv = IPID_table[IPID].pv;
    return S_OK;
  }
  /* find interface (creating proxies as necessary) */
  return COM_GetProxy(OXID, OID, IPID, riid, ppv);
}

static HRESULT WINAPI StdMarshal_ReleaseMarshalData(LPMARSHAL iface,
						    LPSTREAM pStm)
{
  ICOM_THIS(StdMarshalImpl,iface);
  TRACE("(%p)->ReleaseMarshalData(%p)\n",This,pStm);
  /* FIXME: decrease master reference count? */
  return S_OK;
}

static HRESULT WINAPI StdMarshal_DisconnectObject(LPMARSHAL iface,
						  DWORD dwReserved)
{
  ICOM_THIS(StdMarshalImpl,iface);
  TRACE("(%p)->DisconnectObject(%ld)\n",This,dwReserved);
  /* FIXME */
  return S_OK;
}

static ICOM_VTABLE(IMarshal) StdMarshalImpl_VTable =
{
  ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
  StdMarshal_QueryInterface,
  StdMarshal_AddRef,
  StdMarshal_Release,
  StdMarshal_GetUnmarshalClass,
  StdMarshal_GetMarshalSizeMax,
  StdMarshal_MarshalInterface,
  StdMarshal_UnmarshalInterface,
  StdMarshal_ReleaseMarshalData,
  StdMarshal_DisconnectObject
};

/******************** API FUNCTIONS ********************/

HRESULT WINAPI CoGetStandardMarshal(REFIID riid,
				    IUnknown *pUnk,
				    DWORD dwDestContext,
				    LPVOID pvDestContext,
				    DWORD mshlFlags,
				    LPMARSHAL *ppMarshal)
{
  StdMarshalImpl *pMarshal;
  pMarshal = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(StdMarshalImpl));
  if (!pMarshal) return E_OUTOFMEMORY;
  pMarshal->lpVtbl = &StdMarshalImpl_VTable;
  pMarshal->ref = 1;
  *ppMarshal = (LPMARSHAL)pMarshal;
  return S_OK;
}

/* FIXME: we should probably use MS's MEOW thing instead */
typedef struct {
  CLSID proxy;
  DWORD dwContext;
  DWORD mshlFlags;
} WINE_MARSHAL;

static HRESULT COM_GetMarshal(REFIID riid,
			      IUnknown *pUnk,
			      DWORD dwDestContext,
			      void *pvDestContext,
			      DWORD mshlFlags,
			      LPMARSHAL *pMarshal)
{
  HRESULT hr;
  hr = IUnknown_QueryInterface(pUnk, &IID_IMarshal, (LPVOID*)pMarshal);
  if (hr == E_NOINTERFACE) {
    TRACE("no custom marshaler found, using standard marshaler\n");
    hr = CoGetStandardMarshal(riid, pUnk, dwDestContext, pvDestContext,
			      mshlFlags, pMarshal);
  }
  return hr;
}


HRESULT WINAPI CoGetMarshalSizeMax(ULONG *pulSize,
				   REFIID riid,
				   IUnknown *pUnk,
				   DWORD dwDestContext,
				   void *pvDestContext,
				   DWORD mshlFlags)
{
  HRESULT hr = S_OK;
  LPMARSHAL pMarshal;

  /* get an IMarshal interface */
  hr = COM_GetMarshal(riid, pUnk, dwDestContext, pvDestContext, mshlFlags, &pMarshal);
  if (FAILED(hr)) return hr;

  /* get marshal size */
  hr = IMarshal_GetMarshalSizeMax(pMarshal, riid, pUnk, dwDestContext, pvDestContext,
				  mshlFlags, pulSize);

  IMarshal_Release(pMarshal);

  return hr;
}

HRESULT WINAPI CoMarshalInterface(IStream *pStm,
				  REFIID riid,
				  IUnknown *pUnk,
				  DWORD dwDestContext,
				  void *pvDestContext,
				  DWORD mshlFlags)
{
  HRESULT hr;
  WINE_MARSHAL data;
  LPMARSHAL pMarshal;
  ULONG count;
  
  /* get an IMarshal interface */
  hr = COM_GetMarshal(riid, pUnk, dwDestContext, pvDestContext, mshlFlags, &pMarshal);
  if (FAILED(hr)) return hr;
      
  /* get proxy CLSID */
  hr = IMarshal_GetUnmarshalClass(pMarshal, riid, pUnk, dwDestContext, pvDestContext,
				  mshlFlags, &data.proxy);
  if (FAILED(hr)) return hr;

  /* write administrative data */
  data.dwContext = dwDestContext;
  data.mshlFlags = mshlFlags;
  hr = IStream_Write(pStm, &data, sizeof(data), &count);
  if (FAILED(hr)) return hr;
  if (count != sizeof(data)) return E_FAIL;

  /* perform the marshaling */
  hr = IMarshal_MarshalInterface(pMarshal, pStm, riid, pUnk, dwDestContext, pvDestContext,
				 mshlFlags);

  IMarshal_Release(pMarshal);

  return hr;
}

static HRESULT COM_GetUnmarshal(IStream *pStm,
				WINE_MARSHAL *data,
				LPMARSHAL *pMarshal)
{
  HRESULT hr;
  ULONG count;

  /* get administrative data */
  hr = IStream_Read(pStm, data, sizeof(*data), &count);
  if (FAILED(hr)) return hr;
  if (count != sizeof(*data)) return E_FAIL;

  if (IsEqualGUID(&data->proxy, &CLSID_DfMarshal)) {
    /* create standard marshaller */
    hr = CoGetStandardMarshal(NULL, NULL, data->dwContext, NULL, data->mshlFlags, pMarshal);
  } else {
    /* create custom marshaller */
    hr = CoCreateInstance(&data->proxy, NULL, data->dwContext, &IID_IMarshal, (LPVOID*)pMarshal);
  }
  return hr;
}

HRESULT WINAPI CoUnmarshalInterface(IStream *pStm,
				    REFIID riid,
				    void **ppv)
{
  HRESULT hr;
  WINE_MARSHAL data;
  LPMARSHAL pMarshal;

  hr = COM_GetUnmarshal(pStm, &data, &pMarshal);
  if (FAILED(hr)) return hr;

  /* perform the unmarshal */
  hr = IMarshal_UnmarshalInterface(pMarshal, pStm, riid, ppv);
  if (FAILED(hr)) return hr;

  /* clean up marshaling resources */
  if ((data.mshlFlags & 3) == MSHLFLAGS_NORMAL) {
    hr = IMarshal_ReleaseMarshalData(pMarshal, pStm);
  }

  IMarshal_Release(pMarshal);

  return hr;
}

HRESULT WINAPI CoReleaseMarshalData(IStream *pStm)
{
  HRESULT hr;
  WINE_MARSHAL data;
  LPMARSHAL pMarshal;

  hr = COM_GetUnmarshal(pStm, &data, &pMarshal);
  if (FAILED(hr)) return hr;

  hr = IMarshal_ReleaseMarshalData(pMarshal, pStm);

  IMarshal_Release(pMarshal);

  return hr;
}

#define FAKE_INTERTHREAD

HRESULT WINAPI CoMarshalInterThreadInterfaceInStream(REFIID riid,
						     LPUNKNOWN pUnk,
						     LPSTREAM *ppStm)
{
  HRESULT hr = CreateStreamOnHGlobal(0, TRUE, ppStm);
  if (SUCCEEDED(hr)) {
    LARGE_INTEGER pos;
#ifdef FAKE_INTERTHREAD
    hr = IStream_Write(*ppStm, &pUnk, sizeof(pUnk), NULL);
    if (SUCCEEDED(hr)) IUnknown_AddRef(pUnk);
    TRACE("<= %p\n", pUnk);
#else
    hr = CoMarshalInterface(*ppStm, riid, pUnk, MSHCTX_INPROC, 0, MSHLFLAGS_NORMAL);
#endif
    /* FIXME: is the seek done on the marshal or unmarshal side? */
    pos.s.LowPart = 0;
    pos.s.HighPart = 0;
    IStream_Seek(*ppStm, pos, 0, NULL);
  }
  return hr;
}

HRESULT WINAPI CoGetInterfaceAndReleaseStream(LPSTREAM pStm,
					      REFIID riid,
					      LPVOID *ppv)
{
#ifdef FAKE_INTERTHREAD
  LPUNKNOWN pUnk;
  HRESULT hr = IStream_Read(pStm, &pUnk, sizeof(pUnk), NULL);
  if (SUCCEEDED(hr)) {
    hr = IUnknown_QueryInterface(pUnk, riid, ppv);
    TRACE("=> %p\n", *ppv);
    IUnknown_Release(pUnk);
  }
#else
  HRESULT hr = CoUnmarshalInterface(pStm, riid, ppv);
#endif
  IStream_Release(pStm);
  return hr;
}

/******************** BUILT-IN P/S MARSHALERS ********************/
/* we need to create an IDL compiler for Wine that can generate
 * most of these automatically */

typedef void WINAPI (*PRPC_STUB_FUNCTION)(LPRPCSTUBBUFFER This,
					  LPRPCCHANNELBUFFER pChannel,
					  PRPC_MESSAGE pMessage,
					  DWORD *pdwStubPhase);

typedef struct {
  ICOM_VTABLE(IRpcStubBuffer) *lpVtbl;
  DWORD ref;
  IID iid;
  LPUNKNOWN pUnkServer;
  PRPC_STUB_FUNCTION *funcs;
} PSStubImpl;

static HRESULT WINAPI PSStub_QueryInterface(LPRPCSTUBBUFFER iface,
					    REFIID riid,
					    LPVOID *obj)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->QueryInterface(%s,%p)\n",This,debugstr_guid(riid),obj);
  if (IsEqualGUID(&IID_IUnknown,riid) ||
      IsEqualGUID(&IID_IRpcStubBuffer,riid)) {
    *obj = This;
    This->ref++;
    return S_OK;
  }
  return E_NOINTERFACE;
}

static ULONG WINAPI PSStub_AddRef(LPRPCSTUBBUFFER iface)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->AddRef()\n",This);
  return ++(This->ref);
}

static ULONG WINAPI PSStub_Release(LPRPCSTUBBUFFER iface)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->Release()\n",This);
  if (!--(This->ref)) {
    DWORD rc = IUnknown_Release(This->pUnkServer);
    TRACE("server=%p, refcount=%ld\n", This->pUnkServer, rc);
    HeapFree(GetProcessHeap(),0,This);
    return 0;
  }
  return This->ref;
}

static HRESULT WINAPI PSStub_Connect(LPRPCSTUBBUFFER iface,
				     LPUNKNOWN lpUnkServer)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->Connect(%p)\n",This,lpUnkServer);
  This->pUnkServer = lpUnkServer;
  return S_OK;
}

static void WINAPI PSStub_Disconnect(LPRPCSTUBBUFFER iface)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->Disconnect()\n",This);
  This->pUnkServer = NULL;
}

static HRESULT WINAPI PSStub_Invoke(LPRPCSTUBBUFFER iface,
				    PRPCOLEMESSAGE pMsg,
				    LPRPCCHANNELBUFFER pChannel)
{
  ICOM_THIS(PSStubImpl,iface);
  DWORD dwPhase = 0;
  ULONG iMethod = pMsg->iMethod;
  TRACE("(%p)->Invoke(%p,%p)\n",This,pMsg,pChannel);
  if (iMethod == (ULONG)-1) iMethod = 0;
  This->funcs[iMethod](iface, pChannel, (PRPC_MESSAGE)pMsg, &dwPhase);
  return S_OK;
}

static LPRPCSTUBBUFFER WINAPI PSStub_IsIIDSupported(LPRPCSTUBBUFFER iface,
						    REFIID riid)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->IsIIDSupported(%s)\n",This,debugstr_guid(riid));
  return IsEqualGUID(&This->iid,riid) ? iface : NULL;
}

static ULONG WINAPI PSStub_CountRefs(LPRPCSTUBBUFFER iface)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->CountRefs()\n",This);
  return This->ref;
}

static HRESULT WINAPI PSStub_DebugServerQueryInterface(LPRPCSTUBBUFFER iface,
						       LPVOID *ppv)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->DebugServerQueryInterface(%p)\n",This,ppv);
  return S_OK;
}

static void WINAPI PSStub_DebugServerRelease(LPRPCSTUBBUFFER iface,
					      LPVOID pv)
{
  ICOM_THIS(PSStubImpl,iface);
  TRACE("(%p)->DebugServerRelease(%p)\n",This,pv);
}

static ICOM_VTABLE(IRpcStubBuffer) PSStub_VTable =
{
  ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
  PSStub_QueryInterface,
  PSStub_AddRef,
  PSStub_Release,
  PSStub_Connect,
  PSStub_Disconnect,
  PSStub_Invoke,
  PSStub_IsIIDSupported,
  PSStub_CountRefs,
  PSStub_DebugServerQueryInterface,
  PSStub_DebugServerRelease
};

void WINAPI GenStub_QueryInterface(LPRPCSTUBBUFFER This,
				   LPRPCCHANNELBUFFER pChannel,
				   PRPC_MESSAGE pMessage,
				   DWORD *pdwPhase)
{
  FIXME("\n");
}

void WINAPI GenStub_AddRef(LPRPCSTUBBUFFER This,
			   LPRPCCHANNELBUFFER pChannel,
			   PRPC_MESSAGE pMessage,
			   DWORD *pdwPhase)
{
  FIXME("\n");
}

void WINAPI GenStub_Release(LPRPCSTUBBUFFER This,
			    LPRPCCHANNELBUFFER pChannel,
			    PRPC_MESSAGE pMessage,
			    DWORD *pdwPhase)
{
  FIXME("\n");
}

typedef struct {
  void *lpVtbl;
  ICOM_VTABLE(IRpcProxyBuffer) *pVtbl;
  DWORD ref;
  IID iid;
  LPUNKNOWN pUnkOuter;
  LPRPCCHANNELBUFFER pChannel;
} PSProxyImpl;

#define VTBL_OFS(impl) (((LPBYTE)(&((impl*)0)->pVtbl)) - (LPBYTE)0)
#define VTBL_UPCAST(impl,iface) (impl*)(((LPBYTE)iface) - VTBL_OFS(impl))

static HRESULT WINAPI PSProxy_QueryInterface(LPRPCPROXYBUFFER iface,
					     REFIID riid,
					     LPVOID *obj)
{
  PSProxyImpl *This = VTBL_UPCAST(PSProxyImpl,iface);
  TRACE("(%p)->QueryInterface(%s,%p)\n",This,debugstr_guid(riid),obj);
  if (IsEqualGUID(&IID_IUnknown,riid) ||
      IsEqualGUID(&This->iid,riid)) {
    *obj = &This->lpVtbl;
    This->ref++;
    return S_OK;
  }
  if (IsEqualGUID(&IID_IRpcProxyBuffer,riid)) {
    *obj = &This->pVtbl;
    This->ref++;
    return S_OK;
  }
  return E_NOINTERFACE;
}

static ULONG WINAPI PSProxy_AddRef(LPRPCPROXYBUFFER iface)
{
  PSProxyImpl *This = VTBL_UPCAST(PSProxyImpl,iface);
  TRACE("(%p)->AddRef()\n",This);
  return ++(This->ref);
}

static ULONG WINAPI PSProxy_Release(LPRPCPROXYBUFFER iface)
{
  PSProxyImpl *This = VTBL_UPCAST(PSProxyImpl,iface);
  TRACE("(%p)->Release()\n",This);
  if (!--(This->ref)) {
    HeapFree(GetProcessHeap(),0,This);
    return 0;
  }
  return This->ref;
}

static HRESULT WINAPI PSProxy_Connect(LPRPCPROXYBUFFER iface,
				      LPRPCCHANNELBUFFER pChannel)
{
  PSProxyImpl *This = VTBL_UPCAST(PSProxyImpl,iface);
  TRACE("(%p)->Connect(%p)\n",This,pChannel);
  This->pChannel = pChannel;
  return S_OK;
}

static VOID WINAPI PSProxy_Disconnect(LPRPCPROXYBUFFER iface)
{
  PSProxyImpl *This = VTBL_UPCAST(PSProxyImpl,iface);
  TRACE("(%p)->Disconnect()\n",This);
  This->pChannel = NULL;
}

static ICOM_VTABLE(IRpcProxyBuffer) PSProxy_VTable =
{
  ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
  PSProxy_QueryInterface,
  PSProxy_AddRef,
  PSProxy_Release,
  PSProxy_Connect,
  PSProxy_Disconnect
};

static HRESULT WINAPI GenProxy_QueryInterface(LPRPCPROXYBUFFER iface,
					      REFIID riid,
					      LPVOID *obj)
{
  ICOM_THIS(PSProxyImpl,iface);
  TRACE("(%p)->QueryInterface(%s,%p)\n",This,debugstr_guid(riid),obj);
  return IUnknown_QueryInterface(This->pUnkOuter,riid,obj);
}

static ULONG WINAPI GenProxy_AddRef(LPRPCPROXYBUFFER iface)
{
  ICOM_THIS(PSProxyImpl,iface);
  TRACE("(%p)->AddRef()\n",This);
  return ++(This->ref);
}

static ULONG WINAPI GenProxy_Release(LPRPCPROXYBUFFER iface)
{
  ICOM_THIS(PSProxyImpl,iface);
  TRACE("(%p)->Release()\n",This);
  if (!--(This->ref)) {
    /* FIXME: do what aggregated objects do? */
    return 0;
  }
  return This->ref;
}

/***** IClassFactory Stub *****/

void WINAPI CFStub_CreateInstance(LPRPCSTUBBUFFER iface,
				  LPRPCCHANNELBUFFER pChannel,
				  PRPC_MESSAGE pMessage,
				  DWORD *pdwPhase)
{
  ICOM_THIS(PSStubImpl,iface);
  IID iid;
  LPVOID pvObject = NULL;
  HRESULT hr;
  TRACE("(...)\n");
  if (pMessage->BufferLength<sizeof(iid)) return;
  memcpy(&iid,pMessage->Buffer,sizeof(iid));
  TRACE("riid=%s\n", debugstr_guid(&iid));
  hr = IClassFactory_CreateInstance((LPCLASSFACTORY)This->pUnkServer, NULL, &iid, &pvObject);
  TRACE("returns %p\n", pvObject);
  if (pvObject) {
    LPSTREAM pStm = NULL;
    STATSTG stat;
    HGLOBAL hGlobal = 0;
    CreateStreamOnHGlobal(0, TRUE, &pStm);
    CoMarshalInterface(pStm, &iid, pvObject, MSHCTX_LOCAL, NULL, MSHLFLAGS_NORMAL);
    IUnknown_Release((LPUNKNOWN)pvObject);
    IStream_Stat(pStm, &stat, 0);
    pMessage->BufferLength = stat.cbSize.s.LowPart;
    pMessage->Buffer = HeapReAlloc(GetProcessHeap(),0,pMessage->Buffer,pMessage->BufferLength);
    GetHGlobalFromStream(pStm, &hGlobal);
    memcpy(pMessage->Buffer, GlobalLock(hGlobal), pMessage->BufferLength);
    GlobalUnlock(hGlobal);
    IStream_Release(pStm);
  } else {
    HeapFree(GetProcessHeap(),0,pMessage->Buffer);
    pMessage->Buffer = NULL;
    pMessage->BufferLength = 0;
  }
  return;
}

void WINAPI CFStub_LockServer(LPRPCSTUBBUFFER iface,
			      LPRPCCHANNELBUFFER pChannel,
			      PRPC_MESSAGE pMessage,
			      DWORD *pdwPhase)
{
  FIXME("\n");
}

PRPC_STUB_FUNCTION CF_funcs[]={
  GenStub_QueryInterface,
  GenStub_AddRef,
  GenStub_Release,
  CFStub_CreateInstance,
  CFStub_LockServer
};

/***** IClassFactory Proxy *****/

static HRESULT WINAPI CFProxy_CreateInstance(LPCLASSFACTORY iface,
					     LPUNKNOWN pUnkOuter,
					     REFIID riid,
					     LPVOID *ppvObject)
{
  ICOM_THIS(PSProxyImpl,iface);
  RPCOLEMESSAGE msg;
  HRESULT hr;
  ULONG status;
  TRACE("(%p)->CreateInstance(%p,%s,%p)\n",This,pUnkOuter,debugstr_guid(riid),ppvObject);
  if (pUnkOuter) {
    FIXME("need to handle pUnkOuter!\n");
    return E_FAIL;
  }
  msg.reserved1 = This; /* BAD hack */
  msg.iMethod = 3;
  msg.cbBuffer = sizeof(*riid);
  msg.Buffer = NULL;
  IRpcChannelBuffer_GetBuffer(This->pChannel, &msg, &IID_IClassFactory);
  memcpy(msg.Buffer, riid, sizeof(*riid));
  hr = IRpcChannelBuffer_SendReceive(This->pChannel, &msg, &status);
  if (FAILED(hr)) return hr;
  if (msg.cbBuffer) {
    LPSTREAM pStm = NULL;
    ULARGE_INTEGER cbSize;
    HGLOBAL hGlobal = 0;
    hr = CreateStreamOnHGlobal(0, TRUE, &pStm);
    cbSize.s.HighPart = 0;
    cbSize.s.LowPart = msg.cbBuffer;
    IStream_SetSize(pStm, cbSize);
    GetHGlobalFromStream(pStm, &hGlobal);
    memcpy(GlobalLock(hGlobal), msg.Buffer, msg.cbBuffer);
    GlobalUnlock(hGlobal);
    hr = CoUnmarshalInterface(pStm, riid, ppvObject);
    IStream_Release(pStm);
    /* HACK: make sure the new proxy is connected */
    if (SUCCEEDED(hr)) {
      COM_ConnectProxy(*ppvObject, This->pChannel);
    } else {
      ERR("unmarshal failed: %lx\n", hr);
    }
    return hr;
  } else *ppvObject = NULL;
  return status;
}

static HRESULT WINAPI CFProxy_LockServer(LPCLASSFACTORY iface,
					BOOL fLock)
{
  FIXME("not work yet\n");
  return E_FAIL;
}

static ICOM_VTABLE(IClassFactory) CFProxyVT =
{
  ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
  GenProxy_QueryInterface,
  GenProxy_AddRef,
  GenProxy_Release,
  CFProxy_CreateInstance,
  CFProxy_LockServer
};

/***** PSFactory *****/

static GUID CLSID_PSFactoryBuffer = {
  0x320, 0, 0, {0xC0, 0, 0, 0, 0, 0, 0, 0x46}
};

static HRESULT WINAPI PSFactory_QueryInterface(LPPSFACTORYBUFFER iface,
					       REFIID riid,
					       LPVOID *obj)
{
  TRACE("(%p)->QueryInterface(%s,%p)\n",iface,debugstr_guid(riid),obj);
  if (IsEqualGUID(&IID_IUnknown,riid) ||
      IsEqualGUID(&IID_IPSFactoryBuffer,riid)) {
    *obj = iface;
    return S_OK;
  }
  return E_NOINTERFACE;
}

static ULONG WINAPI PSFactory_AddRef(LPPSFACTORYBUFFER iface)
{
  TRACE("(%p)->AddRef()\n",iface);
  return 2;
}

static ULONG WINAPI PSFactory_Release(LPPSFACTORYBUFFER iface)
{
  TRACE("(%p)->Release()\n",iface);
  return 1;
}

static HRESULT WINAPI PSFactory_CreateProxy(LPPSFACTORYBUFFER iface,
					    LPUNKNOWN pUnkOuter,
					    REFIID riid,
					    LPRPCPROXYBUFFER *ppProxy,
					    LPVOID *ppv)
{
  PSProxyImpl *proxy;
  TRACE("(%p)->CreateProxy(%p,%s,%p,%p)\n",iface,pUnkOuter,
	debugstr_guid(riid),ppProxy,ppv);
  if (IsEqualGUID(riid, &IID_IClassFactory)) {
    proxy = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(PSProxyImpl));
    if (!proxy) return E_OUTOFMEMORY;
    proxy->pVtbl = &PSProxy_VTable;
    proxy->ref = 1;
    memcpy(&proxy->iid, riid, sizeof(proxy->iid));
    proxy->pUnkOuter = pUnkOuter;
    proxy->lpVtbl = &CFProxyVT;
    *ppProxy = (LPRPCPROXYBUFFER)&proxy->pVtbl;
    *ppv = proxy;
    TRACE("created proxy for IClassFactory\n");
    return S_OK;
  }
  return E_FAIL;
}

static HRESULT WINAPI PSFactory_CreateStub(LPPSFACTORYBUFFER iface,
					   REFIID riid,
					   LPUNKNOWN pUnkServer,
					   LPRPCSTUBBUFFER *ppStub)
{
  PSStubImpl *stub;
  TRACE("(%p)->CreateStub(%s,%p,%p)\n",iface,debugstr_guid(riid),
	pUnkServer,ppStub);
  if (IsEqualGUID(riid, &IID_IClassFactory)) {
    stub = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(PSStubImpl));
    if (!stub) return E_OUTOFMEMORY;
    stub->lpVtbl = &PSStub_VTable;
    stub->ref = 1;
    memcpy(&stub->iid, riid, sizeof(stub->iid));
    IUnknown_AddRef(pUnkServer);
    stub->pUnkServer = pUnkServer;
    stub->funcs = CF_funcs;
    *ppStub = (LPRPCSTUBBUFFER)stub;
    TRACE("created stub for IClassFactory\n");
    return S_OK;
  }
  return E_FAIL;
}

static ICOM_VTABLE(IPSFactoryBuffer) PSFactoryBuffer_VTable =
{
  ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
  PSFactory_QueryInterface,
  PSFactory_AddRef,
  PSFactory_Release,
  PSFactory_CreateProxy,
  PSFactory_CreateStub
};

static ICOM_VTABLE(IPSFactoryBuffer) *PSFactoryBuffer = &PSFactoryBuffer_VTable;

/***********************************************************************
 *           DllGetClassObject [OLE32.63]
 */
HRESULT WINAPI OLE32_DllGetClassObject(REFCLSID rclsid, REFIID iid,LPVOID *ppv)
{       
  *ppv = NULL;
  if (IsEqualGUID(rclsid, &CLSID_PSFactoryBuffer) &&
      IsEqualGUID(iid, &IID_IPSFactoryBuffer)) {
    *ppv = &PSFactoryBuffer;
    return S_OK;
  }
  FIXME("\n\tCLSID:\t%s,\n\tIID:\t%s\n",debugstr_guid(rclsid),debugstr_guid(iid));
  return CLASS_E_CLASSNOTAVAILABLE;
}
